#include <limits.h> //ULONG_MAX
#include <math.h> //round()
#include "ntplib.h"

/* The library performs the basic arithmetic operation between NTP timestamps, and the
 * conversion to/from local clock values, in milleseconds. Beware of overflow and casting 
 * problems. Fuctions detect and notify some problems.
 */

/* Computes the sum of the first two parameters, and places in the third parameter.
 * In case of overflow, the third parameter is meaningless, and the return value is false
 * One marginal possibility of overflow when the sum is exactly 2^32-1 and there is a carry 
 * in the fractional part.
 */
bool NTPsum(NTPtime *op1, NTPtime *op2, NTPtime *s) {
  s->sec = op1->sec + op2->sec;
  if ((op1->sec) > s->sec || (op2->sec) > s->sec) return false;
  s->frac = op1->frac + op2->frac;
  if ((op1->frac) > s->frac || (op2->frac) > s->frac) s->sec++;
  return true;
}

/* Computes the difference between the first and the second parameter, and places it in
 * the third one. The value is always positive, and the sign of the difference is returned
 * as the result.
 */
int NTPdiff(NTPtime *t0, NTPtime *t1, NTPtime *d) { 
  int sign=1;
  NTPtime *a=t0;
  NTPtime *b=t1;
  if ((t1->sec>t0->sec)||((t1->sec==t0->sec)&&(t1->frac>t0->frac))) {
    sign=-1;
    a=t1;
    b=t0;
  }
  if (a->frac>b->frac) {
    d->frac = a->frac - b->frac;
    d->sec = a->sec - b->sec;
  } else {
    d->frac = (a->frac) + ~(b->frac);
    d->sec = a->sec - b->sec - 1;
  }
  return sign;
}

/* Converts from msecs in an unsigned long integer to NTP notation */
void ms2NTP(NTPtime *dest, unsigned long ms) {
  dest->sec=ms/1000;
  dest->frac=(((ms%1000)<<12)/1000)<<20;
}

/* Converts from NTP 64 bit fixed point representation in the parameter to msec and returs
 *  the result. In case of overflow returns 0
 */
unsigned long NTP2ms(NTPtime *ntp) {
  unsigned long ms;
  if ( ntp->sec > (ULONG_MAX/1000) ) return 0;
  ms=(((ntp->frac)>>20)*1000)>>11;
  ms=round(ms/2.0); //round last digit
  return (ntp->sec)*1000+ms;
}

/* Prints an NTP time as the integer part plus the first prec digits of the fractional part
 * in base 10 representation, with "prec" digits after the decimal dot.
 */
int printNTP(NTPtime *t, const char *name, int prec) {
  Serial.print(name); 
  Serial.print("="); 
  Serial.print (t->sec); 
  Serial.print(" + ");
  Serial.println(((double)(t->frac>>18)/(1<<14)),prec);
  return 0;
}


